"""CLI module for parsing pbf file to geoparquet."""

import json
import logging
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, Optional, Union, cast

import click
import typer
from rq_geo_toolkit._geopandas_api_version import GEOPANDAS_NEW_API

from quackosm._constants import (
    PARQUET_COMPRESSION,
    PARQUET_COMPRESSION_LEVEL,
    PARQUET_ROW_GROUP_SIZE,
    PARQUET_VERSION,
)
from quackosm._osm_tags_filters import GroupedOsmTagsFilter, OsmTagsFilter
from quackosm.osm_extracts.extract import OsmExtractSource
from quackosm.pbf_file_reader import _is_url_path

if TYPE_CHECKING:
    from typing import Literal

    from quackosm._rich_progress import VERBOSITY_MODE

app = typer.Typer(context_settings={"help_option_names": ["-h", "--help"]}, rich_markup_mode="rich")


def _version_callback(value: bool) -> None:
    if value:
        from quackosm import __app_name__, __version__

        typer.echo(f"{__app_name__} {__version__}")
        raise typer.Exit()


def _display_osm_extracts_callback(ctx: typer.Context, value: bool) -> None:
    if value:
        from quackosm.osm_extracts import display_available_extracts

        param_values = {p.name: p.default for p in ctx.command.params}
        param_values.update(ctx.params)
        osm_source = cast("str", param_values.get("osm_extract_source"))
        display_available_extracts(source=osm_source, use_full_names=True, use_pager=True)
        raise typer.Exit()


def _path_callback(ctx: typer.Context, value: Path) -> Path:
    if not _is_url_path(value) and not Path(value).exists():
        raise typer.BadParameter(f"File not found error: {value}")
    return value


def _empty_path_callback(ctx: typer.Context, value: Path) -> Optional[Path]:
    if not value:
        return None
    return _path_callback(ctx, value)


class BboxGeometryParser(click.ParamType):  # type: ignore
    """Parser for geometry in WKT form."""

    name = "BBOX"

    def convert(self, value, param=None, ctx=None):  # type: ignore
        """Convert parameter value."""
        try:
            from shapely import box

            bbox_values = [float(x.strip()) for x in value.split(",")]
            return box(*bbox_values)
        except ValueError:  # ValueError raised when passing non-numbers to float()
            raise typer.BadParameter(
                "Cannot parse provided bounding box."
                " Valid value must contain 4 floating point numbers"
                " separated by commas."
            ) from None


class WktGeometryParser(click.ParamType):  # type: ignore
    """Parser for geometry in WKT form."""

    name = "TEXT (WKT)"

    def convert(self, value, param=None, ctx=None):  # type: ignore
        """Convert parameter value."""
        if not value:
            return None
        try:
            from shapely import from_wkt

            return from_wkt(value)
        except Exception:
            raise typer.BadParameter("Cannot parse provided WKT") from None


class GeoJsonGeometryParser(click.ParamType):  # type: ignore
    """Parser for geometry in GeoJSON form."""

    name = "TEXT (GeoJSON)"

    def convert(self, value, param=None, ctx=None):  # type: ignore
        """Convert parameter value."""
        if not value:
            return None
        try:
            from shapely import from_geojson

            return from_geojson(value)
        except Exception:
            raise typer.BadParameter("Cannot parse provided GeoJSON") from None


class GeoFileGeometryParser(click.ParamType):  # type: ignore
    """Parser for geometry in geo file form."""

    name = "PATH"

    def convert(self, value, param=None, ctx=None):  # type: ignore
        """Convert parameter value."""
        if not value:
            return None

        value = _path_callback(ctx=ctx, value=value)

        try:
            import geopandas as gpd

            gdf = gpd.read_file(value)
            if GEOPANDAS_NEW_API:
                return gdf.union_all()
            else:
                return gdf.unary_union
        except Exception:
            raise typer.BadParameter("Cannot parse provided geo file") from None


class GeocodeGeometryParser(click.ParamType):  # type: ignore
    """Parser for geometry in string Nominatim query form."""

    name = "TEXT"

    def convert(self, value, param=None, ctx=None):  # type: ignore
        """Convert parameter value."""
        if not value:
            return None

        try:
            from quackosm.geocode import geocode_to_geometry

            return geocode_to_geometry(value)
        except Exception:
            raise typer.BadParameter("Cannot geocode provided Nominatim query") from None


class GeohashGeometryParser(click.ParamType):  # type: ignore
    """Parser for geometry in string Nominatim query form."""

    name = "TEXT (Geohash)"

    def convert(self, value, param=None, ctx=None):  # type: ignore
        """Convert parameter value."""
        if not value:
            return None

        try:
            import geopandas as gpd
            from shapely.geometry import box

            from quackosm._geohash_parser import geohash_bounds

            geometries = []
            for geohash in value.split(","):
                bounds = geohash_bounds(geohash.strip())
                geometries.append(box(*bounds))
            if GEOPANDAS_NEW_API:
                return gpd.GeoSeries(geometries).union_all()
            else:
                return gpd.GeoSeries(geometries).unary_union
        except Exception:
            raise


class H3GeometryParser(click.ParamType):  # type: ignore
    """Parser for geometry in string Nominatim query form."""

    name = "TEXT (H3)"

    def convert(self, value, param=None, ctx=None):  # type: ignore
        """Convert parameter value."""
        if not value:
            return None

        import duckdb
        import geopandas as gpd
        from shapely import from_wkt

        duckdb.install_extension("h3", repository="community")
        duckdb.load_extension("h3")

        geometries = []  # noqa: FURB138
        for h3_index in value.split(","):
            stripped_h3_index = h3_index.strip()
            if not duckdb.sql(f"SELECT h3_is_valid_cell('{stripped_h3_index}')").fetchone()[0]:
                raise typer.BadParameter(
                    f"Cannot parse provided H3 value: {stripped_h3_index}"
                ) from None

            parsed_geometry = from_wkt(
                duckdb.sql(f"SELECT h3_cell_to_boundary_wkt('{stripped_h3_index}')").fetchone()[0]
            )

            geometries.append(parsed_geometry)

        if not GEOPANDAS_NEW_API:
            return gpd.GeoSeries(geometries).unary_union

        return gpd.GeoSeries(geometries).union_all()


class S2GeometryParser(click.ParamType):  # type: ignore
    """Parser for geometry in string Nominatim query form."""

    name = "TEXT (S2)"

    def convert(self, value, param=None, ctx=None):  # type: ignore
        """Convert parameter value."""
        if not value:
            return None

        import geopandas as gpd
        import s2sphere
        from shapely import Polygon

        geometries = []  # noqa: FURB138
        for s2_index in value.split(","):
            stripped_s2_index = s2_index.strip()
            try:
                s2_cell = s2sphere.Cell(s2sphere.CellId.from_token(stripped_s2_index))
                points = [
                    s2sphere.LatLng.from_point(s2_cell.get_vertex(i)) for i in (0, 1, 2, 3, 0)
                ]
                geometries.append(
                    Polygon([[point.lng().degrees, point.lat().degrees] for point in points])
                )
            except Exception:
                raise typer.BadParameter(
                    f"Cannot parse provided S2 value: {stripped_s2_index}"
                ) from None

        if not GEOPANDAS_NEW_API:
            return gpd.GeoSeries(geometries).unary_union

        return gpd.GeoSeries(geometries).union_all()


class OsmTagsFilterJsonParser(click.ParamType):  # type: ignore
    """Parser for OSM tags filter in JSON form."""

    name = "TEXT (JSON)"

    def convert(self, value, param=None, ctx=None):  # type: ignore
        """Convert parameter value."""
        if not value:
            return None
        try:
            parsed_dict = json.loads(value)
        except Exception:
            raise typer.BadParameter("Cannot parse provided OSM tags filter") from None

        from quackosm._typing import is_expected_type

        if not is_expected_type(parsed_dict, OsmTagsFilter) and not is_expected_type(
            parsed_dict, GroupedOsmTagsFilter
        ):
            raise typer.BadParameter(
                "Provided OSM tags filter is not in a required format."
            ) from None

        return cast("Union[OsmTagsFilter, GroupedOsmTagsFilter]", parsed_dict)


class OsmTagsFilterFileParser(OsmTagsFilterJsonParser):
    """Parser for OSM tags filter in file form."""

    name = "PATH"

    def convert(self, value, param=None, ctx=None):  # type: ignore
        """Convert parameter value."""
        if not value:
            return None

        value = _path_callback(ctx=ctx, value=value)

        return super().convert(Path(value).read_text(), param, ctx)  # type: ignore


def _filter_osm_ids_callback(value: str) -> Optional[list[str]]:
    if not value:
        return None

    import re

    osm_ids = value.split(",")
    matcher = re.compile(r"^(node|way|relation)\/\d*$")
    parsed_osm_ids = []
    for osm_id in osm_ids:
        stripped_osm_id = osm_id.strip()
        if not stripped_osm_id.startswith(("node/", "way/", "relation/")) or not matcher.match(
            stripped_osm_id
        ):
            raise typer.BadParameter(f"Cannot parse provided OSM id: {stripped_osm_id}") from None
        parsed_osm_ids.append(stripped_osm_id)

    return parsed_osm_ids


@app.command()  # type: ignore
def main(
    pbf_file: Annotated[
        Optional[str],
        typer.Argument(
            help="PBF file to convert into GeoParquet. Can be an URL.",
            metavar="PBF file path",
            callback=_empty_path_callback,
            show_default=False,
        ),
    ] = None,
    osm_tags_filter: Annotated[
        Optional[str],
        typer.Option(
            help=(
                "OSM tags used to filter the data in the "
                "[bold dark_orange]JSON text[/bold dark_orange] form."
                " Can take the form of a flat or grouped dict "
                "(look: [bold green]OsmTagsFilter[/bold green]"
                " and [bold green]GroupedOsmTagsFilter[/bold green])."
                " Cannot be used together with"
                " [bold bright_cyan]osm-tags-filter-file[/bold bright_cyan]."
            ),
            click_type=OsmTagsFilterJsonParser(),
            show_default=False,
        ),
    ] = None,
    osm_tags_filter_file: Annotated[
        Optional[str],
        typer.Option(
            help=(
                "OSM tags used to filter the data in the "
                "[bold dark_orange]JSON file[/bold dark_orange] form."
                " Can take the form of a flat or grouped dict "
                "(look: [bold green]OsmTagsFilter[/bold green]"
                " and [bold green]GroupedOsmTagsFilter[/bold green])."
                " Cannot be used together with"
                " [bold bright_cyan]osm-tags-filter[/bold bright_cyan]."
            ),
            click_type=OsmTagsFilterFileParser(),
            show_default=False,
        ),
    ] = None,
    keep_all_tags: Annotated[
        bool,
        typer.Option(
            "--keep-all-tags/",
            "--all-tags/",
            help=(
                "Whether to keep all tags while filtering with OSM tags."
                " Doesn't work when there is no OSM tags filter applied"
                " ([bold bright_cyan]osm-tags-filter[/bold bright_cyan]"
                " or [bold bright_cyan]osm-tags-filter-file[/bold bright_cyan])."
                " Will override grouping if [bold green]GroupedOsmTagsFilter[/bold green]"
                " has been passed as a filter."
            ),
            show_default=False,
        ),
    ] = False,
    geom_filter_bbox: Annotated[
        Optional[str],
        typer.Option(
            help=(
                "Geometry to use as a filter in the"
                " [bold dark_orange]bounding box[/bold dark_orange] format - 4 floating point"
                " numbers separated by commas."
                " Cannot be used together with other"
                " [bold bright_cyan]geom-filter-...[/bold bright_cyan] parameters."
            ),
            click_type=BboxGeometryParser(),
            show_default=False,
        ),
    ] = None,
    geom_filter_file: Annotated[
        Optional[str],
        typer.Option(
            help=(
                "Geometry to use as a filter in the"
                " [bold dark_orange]file[/bold dark_orange] format - any that can be opened by"
                " GeoPandas. Will return the unary union of the geometries in the file."
                " Cannot be used together with other"
                " [bold bright_cyan]geom-filter-...[/bold bright_cyan] parameters."
            ),
            click_type=GeoFileGeometryParser(),
            show_default=False,
        ),
    ] = None,
    geom_filter_geocode: Annotated[
        Optional[str],
        typer.Option(
            help=(
                "Geometry to use as a filter in the"
                " [bold dark_orange]string to geocode[/bold dark_orange] format - it will be"
                " geocoded to the geometry using Nominatim API (GeoPy library)."
                " Cannot be used together with other"
                " [bold bright_cyan]geom-filter-...[/bold bright_cyan] parameters."
            ),
            click_type=GeocodeGeometryParser(),
            show_default=False,
        ),
    ] = None,
    geom_filter_geojson: Annotated[
        Optional[str],
        typer.Option(
            help=(
                "Geometry to use as a filter in the [bold dark_orange]GeoJSON[/bold dark_orange]"
                " format."
                " Cannot be used together with other"
                " [bold bright_cyan]geom-filter-...[/bold bright_cyan] parameters."
            ),
            click_type=GeoJsonGeometryParser(),
            show_default=False,
        ),
    ] = None,
    geom_filter_index_geohash: Annotated[
        Optional[str],
        typer.Option(
            help=(
                "Geometry to use as a filter in the"
                " [bold dark_orange]Geohash index[/bold dark_orange]"
                " format. Separate multiple values with a comma."
                " Cannot be used together with other"
                " [bold bright_cyan]geom-filter-...[/bold bright_cyan] parameters."
            ),
            click_type=GeohashGeometryParser(),
            show_default=False,
        ),
    ] = None,
    geom_filter_index_h3: Annotated[
        Optional[str],
        typer.Option(
            help=(
                "Geometry to use as a filter in the [bold dark_orange]H3 index[/bold dark_orange]"
                " format. Separate multiple values with a comma."
                " Cannot be used together with other"
                " [bold bright_cyan]geom-filter-...[/bold bright_cyan] parameters."
            ),
            click_type=H3GeometryParser(),
            show_default=False,
        ),
    ] = None,
    geom_filter_index_s2: Annotated[
        Optional[str],
        typer.Option(
            help=(
                "Geometry to use as a filter in the [bold dark_orange]S2 index[/bold dark_orange]"
                " format. Separate multiple values with a comma."
                " Cannot be used together with other"
                " [bold bright_cyan]geom-filter-...[/bold bright_cyan] parameters."
            ),
            click_type=S2GeometryParser(),
            show_default=False,
        ),
    ] = None,
    geom_filter_wkt: Annotated[
        Optional[str],
        typer.Option(
            help=(
                "Geometry to use as a filter in the [bold dark_orange]WKT[/bold dark_orange]"
                " format."
                " Cannot be used together with other"
                " [bold bright_cyan]geom-filter-...[/bold bright_cyan] parameters."
            ),
            click_type=WktGeometryParser(),
            show_default=False,
        ),
    ] = None,
    custom_sql_filter: Annotated[
        Optional[str],
        typer.Option(
            help=(
                "Allows users to pass custom SQL conditions used to filter OSM features. "
                "It will be embedded into predefined queries and requires DuckDB syntax to operate "
                "on tags map object."
            ),
            case_sensitive=False,
            show_default=False,
        ),
    ] = None,
    osm_extract_query: Annotated[
        Optional[str],
        typer.Option(
            help=(
                "Query to find an OpenStreetMap extract from available sources. "
                "Will automatically find and download OSM extract. "
                "Can be used instead of [bold yellow]PBF file path[/bold yellow] argument."
            ),
            case_sensitive=False,
            show_default=False,
        ),
    ] = None,
    osm_extract_source: Annotated[
        OsmExtractSource,
        typer.Option(
            "--osm-extract-source",
            "--pbf-download-source",
            help=(
                "Source where to download the PBF file from."
                " Can be Geofabrik, BBBike, OSMfr (OpenStreetMap.fr) or any."
            ),
            case_sensitive=False,
            show_default="any",
            is_eager=True,
        ),
    ] = OsmExtractSource.any,
    explode_tags: Annotated[
        Optional[bool],
        typer.Option(
            "--explode-tags/--compact-tags",
            "--explode/--compact",
            help=(
                "Whether to split tags into columns based on the OSM tag keys."
                " If [bold violet]None[/bold violet], it will be set based on"
                " the [bold bright_cyan]osm-tags-filter[/bold bright_cyan]"
                "/[bold bright_cyan]osm-tags-filter-file[/bold bright_cyan]"
                " and [bold bright_cyan]keep-all-tags[/bold bright_cyan] parameters."
                " If there is a tags filter applied without"
                " [bold bright_cyan]keep-all-tags[/bold bright_cyan] then it'll be set to"
                " [bold bright_cyan]explode-tags[/bold bright_cyan]"
                " ([bold green]True[/bold green])."
                " Otherwise it'll be set to [bold magenta]compact-tags[/bold magenta]"
                " ([bold red]False[/bold red])."
            ),
            show_default=None,
        ),
    ] = None,
    result_file_path: Annotated[
        Optional[Path],
        typer.Option(
            "--output",
            "-o",
            help=(
                "Path where to save final result file. If not provided, it will be generated"
                " automatically based on the input pbf file name."
                " Can be [bold green].parquet[/bold green] or"
                " [bold green].db[/bold green] or [bold green].duckdb[/bold green] extension."
            ),
            show_default=False,
        ),
    ] = None,
    duckdb: Annotated[
        bool,
        typer.Option(
            "--duckdb",
            help=(
                "Export to duckdb database. If not provided, data can still be exported if"
                " [bold bright_cyan]output[/bold bright_cyan] has [bold green].db[/bold green]"
                " or [bold green].duckdb[/bold green] extension."
            ),
        ),
    ] = False,
    duckdb_table_name: Annotated[
        Optional[str],
        typer.Option(
            "--duckdb-table-name",
            help="Table name which the data will be imported into in the DuckDB database.",
        ),
    ] = "quackosm",
    compression: Annotated[
        str,
        typer.Option(
            "--compression",
            help="Compression of the final parquet file.",
            show_default=True,
        ),
    ] = PARQUET_COMPRESSION,
    compression_level: Annotated[
        int,
        typer.Option(
            "--compression-level",
            help=(
                "Compression level of the final parquet file. Supported only for zstd compression."
            ),
            show_default=True,
        ),
    ] = PARQUET_COMPRESSION_LEVEL,
    row_group_size: Annotated[
        int,
        typer.Option(
            "--row-group-size",
            help="Approximate number of rows per row group in the final parquet file.",
            show_default=True,
        ),
    ] = PARQUET_ROW_GROUP_SIZE,
    parquet_version: Annotated[
        str,
        typer.Option(
            "--parquet-version",
            help="Type of parquet version used to save final file. Supported options: v1 and v2.",
            show_default=True,
        ),
    ] = PARQUET_VERSION,
    ignore_cache: Annotated[
        bool,
        typer.Option(
            "--ignore-cache/",
            "--no-cache/",
            help="Whether to ignore previously precalculated geoparquet files or not.",
            show_default=False,
        ),
    ] = False,
    working_directory: Annotated[
        Path,
        typer.Option(
            "--working-directory",
            "--work-dir",
            help=(
                "Directory where to save the parsed parquet and geoparquet files."
                " Will be created if doesn't exist."
            ),
        ),
    ] = "files",  # type: ignore
    osm_way_polygon_features_config: Annotated[
        Optional[Path],
        typer.Option(
            "--osm-way-polygon-config",
            help=(
                "Config where alternative OSM way polygon features config is defined."
                " Will determine how to parse way features based on tags."
                " Option is intended for experienced users. It's recommended to disable"
                " cache ([bold bright_cyan]no-cache[/bold bright_cyan]) when using this option,"
                " since file names don't contain information what config file has been used"
                " for file generation."
            ),
            callback=_empty_path_callback,
            show_default=False,
        ),
    ] = None,
    filter_osm_ids: Annotated[
        Optional[str],
        typer.Option(
            "--filter-osm-ids",
            help=(
                "List of OSM features IDs to read from the file."
                " Have to be in the form of 'node/<id>', 'way/<id>' or 'relation/<id>'."
                " Separate multiple values with a comma."
            ),
            callback=_filter_osm_ids_callback,
            show_default=False,
        ),
    ] = None,
    sort_result: Annotated[
        bool,
        typer.Option(
            "--sort/--no-sort",
            help="Whether to sort the final geoparquet file by geometry or not.",
            show_default=True,
        ),
    ] = True,
    ignore_metadata_tags: Annotated[
        bool,
        typer.Option(
            "--ignore-metadata-tags/--keep-metadata-tags",
            help="Whether to remove metadata tags, based on the default GDAL config.",
            show_default=True,
        ),
    ] = True,
    wkt_result: Annotated[
        bool,
        typer.Option(
            "--wkt-result/",
            "--wkt/",
            help="Whether to save the geometry as a WKT string instead of WKB blob.",
            show_default=False,
        ),
    ] = False,
    silent_mode: Annotated[
        bool,
        typer.Option(
            "--silent/",
            help="Whether to disable progress reporting.",
            show_default=False,
        ),
    ] = False,
    transient_mode: Annotated[
        bool,
        typer.Option(
            "--transient/",
            help="Whether to make more transient (concise) progress reporting.",
            show_default=False,
        ),
    ] = False,
    geometry_coverage_iou_threshold: Annotated[
        float,
        typer.Option(
            "--iou-threshold",
            help=(
                "Minimal value of the Intersection over Union metric for selecting the matching OSM"
                " extracts. Is best matching extract has value lower than the threshold, it is"
                " discarded (except the first one). Has to be in range between 0 and 1."
                " Value of 0 will allow every intersected extract, value of 1 will only allow"
                " extracts that match the geometry exactly. Works only when PbfFileReader is asked"
                " to download OSM extracts automatically."
            ),
            show_default=0.01,
            min=0,
            max=1,
        ),
    ] = 0.01,
    allow_uncovered_geometry: Annotated[
        bool,
        typer.Option(
            "--allow-uncovered-geometry/",
            help=(
                "Suppresses an error if some geometry parts aren't covered by any OSM extract."
                " Works only when PbfFileReader is asked to download OSM extracts automatically."
            ),
            show_default=False,
        ),
    ] = False,
    cpu_limit: Annotated[
        Optional[int],
        typer.Option(
            "--cpu-limit",
            help=(
                "Max number of threads available for processing."
                " By default, will use all available threads."
            ),
            show_default=False,
        ),
    ] = None,
    show_extracts: Annotated[
        Optional[bool],
        typer.Option(
            "--show-extracts",
            "--show-osm-extracts",
            help="Show available OSM extracts and exit.",
            callback=_display_osm_extracts_callback,
            is_eager=False,
        ),
    ] = None,
    version: Annotated[
        Optional[bool],
        typer.Option(
            "--version",
            "-v",
            help="Show the application's version and exit.",
            callback=_version_callback,
            is_eager=True,
        ),
    ] = None,
) -> None:
    """
    QuackOSM CLI.

    Wraps convert_pbf_to_parquet, convert_geometry_to_parquet and convert_osm_extract_to_parquet
    functions and prints final path to the saved geoparquet file at the end.
    """
    if parquet_version not in ("v1", "v2"):
        raise typer.BadParameter(
            f"Provided incompatible parquet_version ({parquet_version}). Valid options: v1 and v2."
        )
    parquet_version = cast('Literal["v1", "v2"]', parquet_version)

    number_of_geometries_provided = sum(
        geom is not None
        for geom in (
            geom_filter_bbox,
            geom_filter_file,
            geom_filter_geocode,
            geom_filter_geojson,
            geom_filter_index_geohash,
            geom_filter_index_h3,
            geom_filter_index_s2,
            geom_filter_wkt,
        )
    )
    if number_of_geometries_provided > 1:
        raise typer.BadParameter("Provided more than one geometry for filtering")

    geometry_filter_value = (
        geom_filter_bbox
        or geom_filter_file
        or geom_filter_geocode
        or geom_filter_geojson
        or geom_filter_index_geohash
        or geom_filter_index_h3
        or geom_filter_index_s2
        or geom_filter_wkt
    )

    if pbf_file is osm_extract_query is geometry_filter_value is None:
        from click import Argument
        from click.exceptions import MissingParameter

        raise MissingParameter(
            message=(
                "QuackOSM requires either the path to the pbf file,"
                " an OSM extract query (--osm-extract-query) or a geometry filter"
                " (one of --geom-filter-bbox, --geom-filter-file, --geom-filter-geocode,"
                " --geom-filter-geojson, --geom-filter-index-geohash,"
                " --geom-filter-index-h3, --geom-filter-index-s2, --geom-filter-wkt)"
                " to download the file automatically. All three cannot be empty at once."
            ),
            param=Argument(["pbf_file"], type=Path, metavar="PBF file path"),
        )

    if osm_tags_filter is not None and osm_tags_filter_file is not None:
        raise typer.BadParameter("Provided more than one osm tags filter parameter")

    if transient_mode and silent_mode:
        raise typer.BadParameter("Cannot pass both silent and transient mode at once.")

    verbosity_mode: VERBOSITY_MODE = "verbose"

    if transient_mode:
        verbosity_mode = "transient"
    elif silent_mode:
        verbosity_mode = "silent"

    if wkt_result and sort_result:
        sort_result = False

    logging.disable(logging.CRITICAL)

    is_duckdb = (result_file_path and result_file_path.suffix in (".duckdb", ".db")) or duckdb

    pbf_file_parquet = pbf_file and not is_duckdb
    pbf_file_duckdb = pbf_file and is_duckdb
    osm_extract_parquet = osm_extract_query and not is_duckdb
    osm_extract_duckdb = osm_extract_query and is_duckdb
    geometry_parquet = not pbf_file and not osm_extract_query and not is_duckdb
    geometry_duckdb = not pbf_file and not osm_extract_query and is_duckdb

    if pbf_file_parquet:
        from quackosm.functions import convert_pbf_to_parquet

        result_path = convert_pbf_to_parquet(
            pbf_path=cast("str", pbf_file),
            tags_filter=osm_tags_filter or osm_tags_filter_file,  # type: ignore
            keep_all_tags=keep_all_tags,
            geometry_filter=geometry_filter_value,
            explode_tags=explode_tags,
            ignore_cache=ignore_cache,
            working_directory=working_directory,
            result_file_path=result_file_path,
            compression=compression,
            compression_level=compression_level,
            row_group_size=row_group_size,
            parquet_version=parquet_version,
            osm_way_polygon_features_config=(
                json.loads(Path(osm_way_polygon_features_config).read_text())
                if osm_way_polygon_features_config
                else None
            ),
            ignore_metadata_tags=ignore_metadata_tags,
            filter_osm_ids=filter_osm_ids,  # type: ignore
            custom_sql_filter=custom_sql_filter,
            sort_result=sort_result,
            save_as_wkt=wkt_result,
            verbosity_mode=verbosity_mode,
            cpu_limit=cpu_limit,
        )
    elif pbf_file_duckdb:
        from quackosm.functions import convert_pbf_to_duckdb

        result_path = convert_pbf_to_duckdb(
            pbf_path=cast("str", pbf_file),
            tags_filter=osm_tags_filter or osm_tags_filter_file,  # type: ignore
            keep_all_tags=keep_all_tags,
            geometry_filter=geometry_filter_value,
            explode_tags=explode_tags,
            ignore_cache=ignore_cache,
            working_directory=working_directory,
            result_file_path=result_file_path,
            compression=compression,
            compression_level=compression_level,
            row_group_size=row_group_size,
            parquet_version=parquet_version,
            osm_way_polygon_features_config=(
                json.loads(Path(osm_way_polygon_features_config).read_text())
                if osm_way_polygon_features_config
                else None
            ),
            ignore_metadata_tags=ignore_metadata_tags,
            filter_osm_ids=filter_osm_ids,  # type: ignore
            custom_sql_filter=custom_sql_filter,
            sort_result=sort_result,
            duckdb_table_name=duckdb_table_name or "quackosm",
            verbosity_mode=verbosity_mode,
            cpu_limit=cpu_limit,
        )
    elif osm_extract_parquet:
        from quackosm._exceptions import OsmExtractSearchError
        from quackosm.functions import convert_osm_extract_to_parquet

        try:
            result_path = convert_osm_extract_to_parquet(
                osm_extract_query=cast("str", osm_extract_query),
                osm_extract_source=osm_extract_source,
                tags_filter=osm_tags_filter or osm_tags_filter_file,  # type: ignore
                keep_all_tags=keep_all_tags,
                geometry_filter=geometry_filter_value,
                explode_tags=explode_tags,
                ignore_cache=ignore_cache,
                working_directory=working_directory,
                result_file_path=result_file_path,
                compression=compression,
                compression_level=compression_level,
                row_group_size=row_group_size,
                parquet_version=parquet_version,
                osm_way_polygon_features_config=(
                    json.loads(Path(osm_way_polygon_features_config).read_text())
                    if osm_way_polygon_features_config
                    else None
                ),
                ignore_metadata_tags=ignore_metadata_tags,
                filter_osm_ids=filter_osm_ids,  # type: ignore
                custom_sql_filter=custom_sql_filter,
                sort_result=sort_result,
                save_as_wkt=wkt_result,
                verbosity_mode=verbosity_mode,
                cpu_limit=cpu_limit,
            )
        except OsmExtractSearchError as ex:
            from rich.console import Console

            err_console = Console(stderr=True)
            err_console.print(ex)
            raise typer.Exit(code=1) from None
    elif osm_extract_duckdb:
        from quackosm._exceptions import OsmExtractSearchError
        from quackosm.functions import convert_osm_extract_to_duckdb

        try:
            result_path = convert_osm_extract_to_duckdb(
                osm_extract_query=cast("str", osm_extract_query),
                osm_extract_source=osm_extract_source,
                tags_filter=osm_tags_filter or osm_tags_filter_file,  # type: ignore
                keep_all_tags=keep_all_tags,
                geometry_filter=geometry_filter_value,
                explode_tags=explode_tags,
                ignore_cache=ignore_cache,
                working_directory=working_directory,
                result_file_path=result_file_path,
                compression=compression,
                compression_level=compression_level,
                row_group_size=row_group_size,
                parquet_version=parquet_version,
                osm_way_polygon_features_config=(
                    json.loads(Path(osm_way_polygon_features_config).read_text())
                    if osm_way_polygon_features_config
                    else None
                ),
                ignore_metadata_tags=ignore_metadata_tags,
                filter_osm_ids=filter_osm_ids,  # type: ignore
                custom_sql_filter=custom_sql_filter,
                sort_result=sort_result,
                duckdb_table_name=duckdb_table_name or "quackosm",
                save_as_wkt=wkt_result,
                verbosity_mode=verbosity_mode,
                cpu_limit=cpu_limit,
            )
        except OsmExtractSearchError as ex:
            from rich.console import Console

            err_console = Console(stderr=True)
            err_console.print(ex)
            raise typer.Exit(code=1) from None
    elif geometry_parquet:
        from quackosm.functions import convert_geometry_to_parquet

        result_path = convert_geometry_to_parquet(
            geometry_filter=geometry_filter_value,
            osm_extract_source=osm_extract_source,
            tags_filter=osm_tags_filter or osm_tags_filter_file,  # type: ignore
            keep_all_tags=keep_all_tags,
            explode_tags=explode_tags,
            ignore_cache=ignore_cache,
            working_directory=working_directory,
            result_file_path=result_file_path,
            compression=compression,
            compression_level=compression_level,
            row_group_size=row_group_size,
            parquet_version=parquet_version,
            osm_way_polygon_features_config=(
                json.loads(Path(osm_way_polygon_features_config).read_text())
                if osm_way_polygon_features_config
                else None
            ),
            ignore_metadata_tags=ignore_metadata_tags,
            filter_osm_ids=filter_osm_ids,  # type: ignore
            custom_sql_filter=custom_sql_filter,
            sort_result=sort_result,
            save_as_wkt=wkt_result,
            verbosity_mode=verbosity_mode,
            geometry_coverage_iou_threshold=geometry_coverage_iou_threshold,
            allow_uncovered_geometry=allow_uncovered_geometry,
            cpu_limit=cpu_limit,
        )
    elif geometry_duckdb:
        from quackosm.functions import convert_geometry_to_duckdb

        result_path = convert_geometry_to_duckdb(
            geometry_filter=geometry_filter_value,
            osm_extract_source=osm_extract_source,
            tags_filter=osm_tags_filter or osm_tags_filter_file,  # type: ignore
            keep_all_tags=keep_all_tags,
            explode_tags=explode_tags,
            ignore_cache=ignore_cache,
            working_directory=working_directory,
            result_file_path=result_file_path,
            compression=compression,
            compression_level=compression_level,
            row_group_size=row_group_size,
            parquet_version=parquet_version,
            osm_way_polygon_features_config=(
                json.loads(Path(osm_way_polygon_features_config).read_text())
                if osm_way_polygon_features_config
                else None
            ),
            ignore_metadata_tags=ignore_metadata_tags,
            filter_osm_ids=filter_osm_ids,  # type: ignore
            custom_sql_filter=custom_sql_filter,
            duckdb_table_name=duckdb_table_name or "quackosm",
            sort_result=sort_result,
            save_as_wkt=wkt_result,
            verbosity_mode=verbosity_mode,
            geometry_coverage_iou_threshold=geometry_coverage_iou_threshold,
            allow_uncovered_geometry=allow_uncovered_geometry,
            cpu_limit=cpu_limit,
        )
    else:
        raise RuntimeError("Unknown operation mode")

    typer.secho(result_path, fg="green")



"""
OpenStreetMap.fr extracts.

This module contains wrapper for publically available OpenStreetMap.fr download server.
"""

import re
from typing import Any, Optional

import geopandas as gpd
import requests
from tqdm import tqdm

from quackosm._rich_progress import FORCE_TERMINAL
from quackosm.osm_extracts._poly_parser import parse_polygon_file
from quackosm.osm_extracts.extract import (
    OpenStreetMapExtract,
    OsmExtractSource,
    extracts_to_geodataframe,
    load_index_decorator,
)

OPENSTREETMAP_FR_POLYGONS_INDEX_URL = "https://download.openstreetmap.fr/polygons"
OPENSTREETMAP_FR_EXTRACTS_INDEX_URL = "https://download.openstreetmap.fr/extracts"
OPENSTREETMAP_FR_INDEX_GDF: Optional[gpd.GeoDataFrame] = None

__all__ = ["_get_openstreetmap_fr_index"]


def _get_openstreetmap_fr_index() -> gpd.GeoDataFrame:
    global OPENSTREETMAP_FR_INDEX_GDF  # noqa: PLW0603

    if OPENSTREETMAP_FR_INDEX_GDF is None:
        OPENSTREETMAP_FR_INDEX_GDF = _load_openstreetmap_fr_index()

    return OPENSTREETMAP_FR_INDEX_GDF


@load_index_decorator(OsmExtractSource.osm_fr)
def _load_openstreetmap_fr_index() -> gpd.GeoDataFrame:  # pragma: no cover
    """
    Load available extracts from OpenStreetMap.fr download service.

    Returns:
        gpd.GeoDataFrame: Extracts index with metadata.
    """
    extracts = []
    with tqdm(disable=FORCE_TERMINAL) as pbar:
        extract_soup_objects = _gather_all_openstreetmap_fr_urls(
            OsmExtractSource.osm_fr.value, "/", pbar
        )
        pbar.set_description(OsmExtractSource.osm_fr.value)
        extracts = _parse_openstreetmap_fr_urls(
            pbar=pbar, extract_soup_objects=extract_soup_objects
        )

    gdf = extracts_to_geodataframe(extracts)

    return gdf


def _gather_all_openstreetmap_fr_urls(
    id_prefix: str, directory_url: str, pbar: tqdm
) -> list[Any]:  # pragma: no cover
    """
    Iterate OpenStreetMap.fr extracts service page.

    Works recursively, by scraping whole available directory.

    Args:
        id_prefix (str): Prefix to be applies to extracts names.
        directory_url (str): Directory URL to load.
        pbar (tqdm): Progress bar.

    Returns:
        list[Any]: List of osm.fr extracts urls objects for further processing.
    """
    from bs4 import BeautifulSoup

    pbar.set_description_str(id_prefix)
    extract_soup_objects = []

    result = requests.get(
        f"{OPENSTREETMAP_FR_EXTRACTS_INDEX_URL}{directory_url}",
        headers={"User-Agent": "QuackOSM Python package (https://github.com/kraina-ai/quackosm)"},
    )
    soup = BeautifulSoup(result.text, "html.parser")

    extracts_urls = soup.find_all(string=re.compile("-latest\\.osm\\.pbf$"))
    pbar.total = (pbar.total or 0) + len(extracts_urls)
    pbar.refresh()
    extract_soup_objects.extend(
        [(extract_url, id_prefix, directory_url) for extract_url in extracts_urls]
    )

    directories = soup.find_all(src="/icons/folder.gif")
    for directory in directories:
        link = directory.find_parent("tr").find("a")
        name = link.text.replace("/", "")
        extract_soup_objects.extend(
            _gather_all_openstreetmap_fr_urls(
                id_prefix=f"{id_prefix}_{name}",
                directory_url=f"{directory_url}{link['href']}",
                pbar=pbar,
            )
        )

    return extract_soup_objects


def _parse_openstreetmap_fr_urls(
    pbar: tqdm, extract_soup_objects: list[Any]
) -> list[OpenStreetMapExtract]:
    extracts = []

    for soup_object, id_prefix, directory_url in extract_soup_objects:
        link = soup_object.find_parent("tr").find("a")
        name = link.text.replace("-latest.osm.pbf", "")
        polygon = parse_polygon_file(
            f"{OPENSTREETMAP_FR_POLYGONS_INDEX_URL}/{directory_url}{name}.poly"
        )
        if polygon is None:
            continue
        extracts.append(
            OpenStreetMapExtract(
                id=f"{id_prefix}_{name}",
                name=name,
                parent=id_prefix,
                url=f"{OPENSTREETMAP_FR_EXTRACTS_INDEX_URL}{directory_url}{link['href']}",
                geometry=polygon,
            )
        )
        pbar.set_description_str(id_prefix)
        pbar.update()

    return extracts



"""
QuackOSM.

QuackOSM is a Python library used for reading pbf (ProtoBuffer) files with OpenStreetMap data using
DuckDB spatial extension without GDAL.
"""

from quackosm.functions import (
    convert_geometry_to_duckdb,
    convert_geometry_to_geodataframe,
    convert_geometry_to_parquet,
    convert_osm_extract_to_duckdb,
    convert_osm_extract_to_geodataframe,
    convert_osm_extract_to_parquet,
    convert_pbf_to_duckdb,
    convert_pbf_to_geodataframe,
    convert_pbf_to_parquet,
)
from quackosm.geocode import geocode_to_geometry
from quackosm.osm_extracts import display_available_extracts
from quackosm.pbf_file_reader import PbfFileReader

__app_name__ = "QuackOSM"
__version__ = "0.16.3"

__all__ = [
    "PbfFileReader",
    "convert_pbf_to_parquet",
    "convert_pbf_to_duckdb",
    "convert_geometry_to_parquet",
    "convert_geometry_to_duckdb",
    "convert_osm_extract_to_parquet",
    "convert_osm_extract_to_duckdb",
    "convert_pbf_to_geodataframe",
    "convert_geometry_to_geodataframe",
    "convert_osm_extract_to_geodataframe",
    "display_available_extracts",
    "geocode_to_geometry",
]



"""Tests for proper URL detection."""

import pytest

from quackosm.pbf_file_reader import _is_url_path


@pytest.mark.parametrize(
    "path,is_url",
    [
        ("D:/a/quackosm/files/seychelles.osm.pbf", False),
        ("http://127.0.0.1:8080/seychelles.osm.pbf", True),
        ("ftp://127.0.0.1:8080/seychelles.osm.pbf", True),
        ("/dev/files/seychelles.osm.pbf", False),
        ("files/seychelles.osm.pbf", False),
        ("/files/seychelles.osm.pbf", False),
        ("./files/seychelles.osm.pbf", False),
        ("../files/seychelles.osm.pbf", False),
        ("http://download.geofabrik.de/africa/seychelles-latest.osm.pbf", True),
    ],
)  # type: ignore
def test_url_parsing(path: str, is_url: bool) -> None:
    """Test if URL detection works."""
    assert _is_url_path(path) == is_url



class EmptyResultWarning(Warning): ...


class GeometryNotCoveredWarning(Warning): ...


class GeometryNotCoveredError(Exception): ...


class InvalidGeometryFilter(Exception): ...


class MultiprocessingRuntimeError(RuntimeError): ...


class OsmExtractIndexOutdatedWarning(Warning): ...


class OsmExtractSearchError(Exception):
    def __init__(self, message: str, matching_full_names: list[str]):
        super().__init__(message)
        self.matching_full_names = matching_full_names


class OsmExtractZeroMatchesError(OsmExtractSearchError): ...


class OsmExtractMultipleMatchesError(OsmExtractSearchError): ...


class QueryNotGeocodedError(ValueError): ...


class MissingOsmCacheWarning(Warning): ...


class OldOsmCacheWarning(Warning): ...



"""Run full country file in a low memory environment."""

import multiprocessing
import os
from contextlib import suppress
from pathlib import Path

import duckdb
import psutil
from pooch import get_logger as get_pooch_logger
from pooch import retrieve
from psutil._common import bytes2human

import quackosm as qosm


def test_country_file() -> None:
    """Test if big file is working in a low memory environment."""
    extract_name = "portugal"
    display_resources()
    files_dir = Path("files")
    file_name = files_dir / f"{extract_name}.osm.pbf"
    logger = get_pooch_logger()
    logger.setLevel("WARNING")
    retrieve(
        f"https://download.geofabrik.de/europe/{extract_name}-latest.osm.pbf",
        fname=f"{extract_name}.osm.pbf",
        path=files_dir,
        progressbar=False,
        known_hash=None,
    )

    qosm.convert_pbf_to_parquet(
        pbf_path=file_name, ignore_cache=True, verbosity_mode="verbose", sort_result=True
    )


def display_resources() -> None:
    """Show available resources."""
    print(
        "CPU count (duckdb):",
        duckdb.sql("SELECT current_setting('threads') AS threads").fetchone()[0],
    )
    print("CPU count (multiprocessing):", multiprocessing.cpu_count())
    print("CPU count (psutil):", psutil.cpu_count())
    print("CPU count (psutil physical):", psutil.cpu_count(logical=False))

    print(
        "Memory (duckdb):",
        duckdb.sql("SELECT current_setting('memory_limit') AS memlimit").fetchone()[0],
    )
    print("Memory (psutil):", bytes2human(psutil.virtual_memory().total))
    print("Memory (psutil, all):", psutil.virtual_memory())

    with suppress(Exception):
        print("CPU affinity (os):", os.sched_getaffinity(0))
        print("CPU affinity (psutil):", psutil.Process().cpu_affinity())
        print("Memory limit (cgroup):", bytes2human(int(open("/sys/fs/cgroup/memory.max").read())))


if __name__ == "__main__":
    test_country_file()



"""Tests for PbfFileReader OSM tags filtering."""

from collections.abc import Iterable
from pathlib import Path
from typing import Union
from unittest import TestCase

import pytest

from quackosm._osm_tags_filters import GroupedOsmTagsFilter, OsmTagsFilter, merge_osm_tags_filter
from quackosm.pbf_file_reader import PbfFileReader
from tests.base.conftest import GEOFABRIK_LAYERS, HEX2VEC_FILTER

ut = TestCase()


@pytest.mark.parametrize(  # type: ignore
    "osm_tags_filter,expected_result_filter",
    [
        ({"tag_a": True}, {"tag_a": True}),
        ({"tag_a": "A"}, {"tag_a": "A"}),
        ({"tag_a": ["A"]}, {"tag_a": ["A"]}),
        ({}, {}),
        ({"group_a": {}}, {}),
        ({"group_a": {"tag_a": True}}, {"tag_a": True}),
        ({"group_a": {"tag_a": "A"}}, {"tag_a": ["A"]}),
        ({"group_a": {"tag_a": ["A"]}}, {"tag_a": ["A"]}),
        ({"group_a": {"tag_a": "A", "tag_b": "B"}}, {"tag_a": ["A"], "tag_b": ["B"]}),
        ({"group_a": {"tag_a": ["A"], "tag_b": ["B"]}}, {"tag_a": ["A"], "tag_b": ["B"]}),
        (
            {
                "group_a": {"tag_a": "A", "tag_b": "B"},
                "group_b": {"tag_a": "A", "tag_b": "B"},
            },
            {"tag_a": ["A"], "tag_b": ["B"]},
        ),
        (
            {
                "group_a": {"tag_a": "A", "tag_b": "B"},
                "group_b": {"tag_c": "C", "tag_d": "D"},
            },
            {"tag_a": ["A"], "tag_b": ["B"], "tag_c": ["C"], "tag_d": ["D"]},
        ),
        (
            {
                "group_a": {"tag_a": "A", "tag_b": "B"},
                "group_b": {"tag_a": "C", "tag_b": "D"},
            },
            {"tag_a": ["A", "C"], "tag_b": ["B", "D"]},
        ),
        (
            {
                "group_a": {"tag_a": "A", "tag_b": "B"},
                "group_b": {"tag_a": ["C", "D"], "tag_b": "E"},
            },
            {"tag_a": ["A", "C", "D"], "tag_b": ["B", "E"]},
        ),
        (
            {
                "group_a": {"tag_a": "A", "tag_b": "B"},
                "group_b": {"tag_a": ["C", "D"], "tag_b": True},
            },
            {"tag_a": ["A", "C", "D"], "tag_b": True},
        ),
        (
            {
                "group_a": {"tag_a": ["A", "C"], "tag_b": ["B", "E"]},
                "group_b": {"tag_a": ["C", "D"], "tag_b": ["B"]},
            },
            {"tag_a": ["A", "C", "D"], "tag_b": ["B", "E"]},
        ),
        ([{"tag_a": True}], {"tag_a": True}),
        ([{"tag_a": "A"}], {"tag_a": ["A"]}),
        ([{"tag_a": ["A"]}], {"tag_a": ["A"]}),
        ([{}], {}),
        ([{"group_a": {}}], {}),
        (
            [{"tag_a": "A", "tag_b": "B"}, {"tag_a": "A", "tag_b": "B"}],
            {"tag_a": ["A"], "tag_b": ["B"]},
        ),
        (
            [
                {
                    "group_a": {"tag_a": "A", "tag_b": "B"},
                    "group_b": {"tag_a": "A", "tag_b": "B"},
                },
                {"tag_a": "A", "tag_b": "B"},
            ],
            {"tag_a": ["A"], "tag_b": ["B"]},
        ),
        (
            [
                {
                    "group_a": {"tag_a": "A", "tag_b": "B"},
                    "group_b": {"tag_a": "A", "tag_b": "B"},
                },
                {
                    "group_a": {"tag_a": "A", "tag_b": "B"},
                    "group_b": {"tag_a": "A", "tag_b": "B"},
                },
            ],
            {"tag_a": ["A"], "tag_b": ["B"]},
        ),
        ([{}, {}], {}),
        ([{"group_a": {}}, {"group_a": {}}], {}),
        ([{"group_a": {}}, {}], {}),
        (
            [{"tag_a": "A", "tag_b": "B"}, {"tag_c": "C", "tag_d": "D"}],
            {"tag_a": ["A"], "tag_b": ["B"], "tag_c": ["C"], "tag_d": ["D"]},
        ),
    ],
)
def test_merging_correct_osm_tags_filters(
    osm_tags_filter: Union[
        GroupedOsmTagsFilter, OsmTagsFilter, Iterable[GroupedOsmTagsFilter], Iterable[OsmTagsFilter]
    ],
    expected_result_filter: OsmTagsFilter,
) -> None:
    """Test merging grouped tags filter into a base osm filter."""
    merged_filters = merge_osm_tags_filter(osm_tags_filter)
    ut.assertDictEqual(expected_result_filter, merged_filters)


@pytest.mark.parametrize(  # type: ignore
    "osm_tags_filter",
    [
        {
            "group_a": {"tag_a": True},
            "group_b": {"tag_a": False},
        },
        {
            "group_a": {"tag_a": ["A"], "tag_b": ["B"]},
            "group_b": {"tag_a": False, "tag_c": ["C"]},
        },
        {
            "group_a": {"tag_a": ["A"], "tag_b": ["B"]},
            "group_b": {"tag_a": ["C", "D"], "tag_b": False},
        },
        [{"tag_a": True}, {"tag_a": False}],
        [{"tag_a": False}, {"tag_a": True}],
        [{"tag_a": "A"}, {"tag_a": False}],
        [{"tag_a": ["A"]}, {"tag_a": False}],
    ],
)
def test_merging_incorrect_osm_tags_filters(
    osm_tags_filter: Union[
        GroupedOsmTagsFilter, OsmTagsFilter, Iterable[GroupedOsmTagsFilter], Iterable[OsmTagsFilter]
    ],
) -> None:
    """Test merging grouped tags filter into a base osm filter."""
    with pytest.raises(ValueError):
        merge_osm_tags_filter(osm_tags_filter)


@pytest.mark.parametrize(  # type: ignore
    "test_file_name,query,expected_result_length,expected_features_columns_length",
    [
        (
            "d17f922ed15e9609013a6b895e1e7af2d49158f03586f2c675d17b760af3452e.osm.pbf",
            None,
            678,
            271,
        ),
        (
            "eb2848d259345ce7dfe8af34fd1ab24503bb0b952e04e872c87c55550fa50fbf.osm.pbf",
            None,
            1,
            22,
        ),
        ("529cdcbb7a3cc103658ef31b39bed24984e421127d319c867edf2f86ff3bb098.osm.pbf", None, 0, 0),
        (
            "d17f922ed15e9609013a6b895e1e7af2d49158f03586f2c675d17b760af3452e.osm.pbf",
            HEX2VEC_FILTER,
            97,
            10,
        ),
        (
            "eb2848d259345ce7dfe8af34fd1ab24503bb0b952e04e872c87c55550fa50fbf.osm.pbf",
            HEX2VEC_FILTER,
            0,
            0,
        ),
        (
            "d17f922ed15e9609013a6b895e1e7af2d49158f03586f2c675d17b760af3452e.osm.pbf",
            GEOFABRIK_LAYERS,
            433,
            22,
        ),
        (
            "eb2848d259345ce7dfe8af34fd1ab24503bb0b952e04e872c87c55550fa50fbf.osm.pbf",
            GEOFABRIK_LAYERS,
            0,
            0,
        ),
    ],
)
def test_pbf_reader(
    test_file_name: str,
    query: OsmTagsFilter,
    expected_result_length: int,
    expected_features_columns_length: int,
) -> None:
    """Test proper files loading in `PbfFileReader`."""
    features_gdf = PbfFileReader(tags_filter=query).convert_pbf_to_geodataframe(
        pbf_path=[Path(__file__).parent.parent / "test_files" / test_file_name],
        explode_tags=True,
        ignore_cache=True,
    )
    assert len(features_gdf) == expected_result_length, (
        f"Mismatched result length ({len(features_gdf)}, {expected_result_length})"
    )
    assert len(features_gdf.columns) == expected_features_columns_length + 1, (
        f"Mismatched columns length ({len(features_gdf.columns)},"
        f" {expected_features_columns_length + 1})"
    )


@pytest.mark.parametrize(  # type: ignore
    "filter_osm_id,osm_tags_filter,keep_all_tags,expected_tags_keys",
    [
        ("way/389888402", {"building": "apartments"}, False, ["building"]),
        (
            "way/389888402",
            {"building": "apartments"},
            True,
            [
                "addr:city",
                "addr:country",
                "addr:housenumber",
                "addr:postcode",
                "addr:street",
                "building",
                "building:levels",
            ],
        ),
        ("way/627022271", {"leisure": "garden"}, False, ["leisure"]),
        (
            "way/627022271",
            {"leisure": "garden"},
            True,
            [
                "addr:country",
                "leisure",
                "name",
            ],
        ),
    ],
)
def test_pbf_reader_proper_tags_reading(
    filter_osm_id: str,
    osm_tags_filter: Union[OsmTagsFilter, GroupedOsmTagsFilter],
    keep_all_tags: bool,
    expected_tags_keys: list[str],
) -> None:
    """Test proper tags reading with filtering on osm_id in `PbfFileReader`."""
    file_name = "monaco.osm.pbf"
    features_gdf = PbfFileReader(tags_filter=osm_tags_filter).convert_pbf_to_geodataframe(
        pbf_path=[Path(__file__).parent.parent / "test_files" / file_name],
        ignore_cache=True,
        filter_osm_ids=[filter_osm_id],
        explode_tags=False,
        keep_all_tags=keep_all_tags,
    )
    assert len(features_gdf) == 1
    returned_tags_keys = list(features_gdf.iloc[0].tags.keys())
    ut.assertListEqual(returned_tags_keys, expected_tags_keys)


@pytest.mark.parametrize(  # type: ignore
    "osm_tags_filter,expected_result_length,expected_top_10_ids,expected_no_columns,expected_top_10_columns",
    [
        (
            {"building": True},
            1283,
            [
                "relation/11384697",
                "relation/11484092",
                "relation/11484093",
                "relation/11484094",
                "relation/11485520",
                "relation/11538023",
                "relation/11546879",
                "relation/1369192",
                "relation/1369193",
                "relation/1369195",
            ],
            2,
            ["building", "geometry"],
        ),
        (
            {"amenity": True, "leisure": True},
            1038,
            [
                "node/10020887517",
                "node/10021298117",
                "node/10021298717",
                "node/10025656390",
                "node/10025843517",
                "node/10025852089",
                "node/10025852090",
                "node/10028051243",
                "node/10068880335",
                "node/10127713363",
            ],
            3,
            ["amenity", "geometry", "leisure"],
        ),
        (
            {"amenity": "parking", "leisure": ["park", "garden"], "office": True},
            157,
            [
                "node/10025656391",
                "node/10025656392",
                "node/10039418191",
                "node/1079045434",
                "node/1079750865",
                "node/11094273773",
                "node/11151020213",
                "node/11252553796",
                "node/1661139827",
                "node/1662764987",
            ],
            4,
            ["amenity", "geometry", "leisure", "office"],
        ),
        (
            {"office": False},
            7896,
            [
                "node/10005045289",
                "node/10020887517",
                "node/10021298117",
                "node/10021298717",
                "node/10025656383",
                "node/10025656390",
                "node/10025656393",
                "node/10025656394",
                "node/10025656395",
                "node/10025843517",
            ],
            747,
            [
                "ISO3166-1:alpha2",
                "ISO3166-2",
                "abandoned:railway",
                "access",
                "access:backward",
                "access:covid19",
                "access:lanes",
                "addr:city",
                "addr:country",
                "addr:district",
            ],
        ),
        (
            {"building": True, "office": False},
            1276,
            [
                "relation/11384697",
                "relation/11484092",
                "relation/11484093",
                "relation/11484094",
                "relation/11485520",
                "relation/11538023",
                "relation/11546879",
                "relation/1369192",
                "relation/1369193",
                "relation/1369195",
            ],
            2,
            ["building", "geometry"],
        ),
        (
            {"name:en": True},
            54,
            [
                "node/10039418191",
                "node/10671507005",
                "node/10674256605",
                "node/1661205490",
                "node/1790048269",
                "node/2838235236",
                "node/2838250945",
                "node/4187925032",
                "node/4416171690",
                "node/4416197079",
            ],
            2,
            ["geometry", "name:en"],
        ),
        (
            {"name:e*": True},
            57,
            [
                "node/10039418191",
                "node/10671507005",
                "node/10674256605",
                "node/1661205490",
                "node/1790048269",
                "node/25258130",
                "node/2838235236",
                "node/2838250945",
                "node/4187925032",
                "node/4416171690",
            ],
            9,
            [
                "geometry",
                "name:ee",
                "name:el",
                "name:en",
                "name:eo",
                "name:es",
                "name:et",
                "name:eu",
                "name:ext",
            ],
        ),
        (
            {"name:*": False},
            7806,
            [
                "node/10005045289",
                "node/10020887517",
                "node/10021298117",
                "node/10021298717",
                "node/10025656383",
                "node/10025656390",
                "node/10025656391",
                "node/10025656392",
                "node/10025656393",
                "node/10025656394",
            ],
            477,
            [
                "ISO3166-2",
                "abandoned:railway",
                "access",
                "access:backward",
                "access:covid19",
                "access:lanes",
                "addr:city",
                "addr:country",
                "addr:district",
                "addr:full",
            ],
        ),
        (
            {"*": True},
            7937,
            [
                "node/10005045289",
                "node/10020887517",
                "node/10021298117",
                "node/10021298717",
                "node/10025656383",
                "node/10025656390",
                "node/10025656391",
                "node/10025656392",
                "node/10025656393",
                "node/10025656394",
            ],
            758,
            [
                "ISO3166-1:alpha2",
                "ISO3166-2",
                "abandoned:railway",
                "access",
                "access:backward",
                "access:covid19",
                "access:lanes",
                "addr:city",
                "addr:country",
                "addr:district",
            ],
        ),
        ({"*": False}, 0, [], 1, ["geometry"]),
        (
            {"building": True, "addr:*": False},
            286,
            [
                "relation/11484092",
                "relation/11485520",
                "relation/11538023",
                "relation/1369192",
                "relation/1369193",
                "relation/1369195",
                "relation/1484217",
                "relation/16248281",
                "relation/16248282",
                "relation/16248283",
            ],
            2,
            ["building", "geometry"],
        ),
        (
            {"building": True, "addr:*": True},
            1935,
            [
                "node/10025843517",
                "node/10025852088",
                "node/10025852090",
                "node/10028051243",
                "node/10039418191",
                "node/10060634234",
                "node/10060634235",
                "node/10060634237",
                "node/10060634238",
                "node/10060634239",
            ],
            10,
            [
                "addr:city",
                "addr:country",
                "addr:district",
                "addr:full",
                "addr:housename",
                "addr:housenumber",
                "addr:postcode",
                "addr:street",
                "building",
                "geometry",
            ],
        ),
        (
            {"building": True, "addr:*": True, "source:*": False},
            1933,
            [
                "node/10025843517",
                "node/10025852088",
                "node/10025852090",
                "node/10028051243",
                "node/10039418191",
                "node/10060634234",
                "node/10060634235",
                "node/10060634237",
                "node/10060634238",
                "node/10060634239",
            ],
            10,
            [
                "addr:city",
                "addr:country",
                "addr:district",
                "addr:full",
                "addr:housename",
                "addr:housenumber",
                "addr:postcode",
                "addr:street",
                "building",
                "geometry",
            ],
        ),
        (
            {"name:*": "Monaco"},
            2,
            ["node/1790048269", "node/6684051501"],
            46,
            [
                "geometry",
                "name:af",
                "name:als",
                "name:ang",
                "name:bar",
                "name:bi",
                "name:cy",
                "name:da",
                "name:de",
                "name:dsb",
            ],
        ),
        (
            {"name:*": ["Monaco", "France"]},
            2,
            ["node/1790048269", "node/6684051501"],
            46,
            [
                "geometry",
                "name:af",
                "name:als",
                "name:ang",
                "name:bar",
                "name:bi",
                "name:cy",
                "name:da",
                "name:de",
                "name:dsb",
            ],
        ),
        (
            {"highway": "primary", "maxspeed": False},
            63,
            [
                "way/1019174499",
                "way/1019360049",
                "way/1019360050",
                "way/1024393367",
                "way/1024393368",
                "way/1131276186",
                "way/1131276187",
                "way/1174006399",
                "way/1174006400",
                "way/1174006401",
            ],
            2,
            ["geometry", "highway"],
        ),
        (
            {"highway": "*ary"},
            400,
            [
                "way/1001561436",
                "way/1001561437",
                "way/1001561438",
                "way/1019174499",
                "way/1019174508",
                "way/1019174509",
                "way/1019360049",
                "way/1019360050",
                "way/1019646822",
                "way/1024393367",
            ],
            2,
            ["geometry", "highway"],
        ),
        (
            {"highway": "*ary", "maxspeed": False},
            198,
            [
                "way/1019174499",
                "way/1019360049",
                "way/1019360050",
                "way/1024393367",
                "way/1024393368",
                "way/1058067231",
                "way/1058067232",
                "way/1099877047",
                "way/1099877048",
                "way/1131276186",
            ],
            2,
            ["geometry", "highway"],
        ),
        (
            {"*speed": "*0"},
            293,
            [
                "node/1079045434",
                "way/1001561436",
                "way/1001561437",
                "way/1001561438",
                "way/1019174508",
                "way/1019174509",
                "way/1019646822",
                "way/1024523794",
                "way/1024523795",
                "way/1024523796",
            ],
            2,
            ["geometry", "maxspeed"],
        ),
        (
            {"*speed": "*0", "railway": False, "waterway": False},
            264,
            [
                "node/1079045434",
                "way/1001561436",
                "way/1001561437",
                "way/1001561438",
                "way/1019174508",
                "way/1019174509",
                "way/1019646822",
                "way/1024523794",
                "way/1024523795",
                "way/1024523796",
            ],
            2,
            ["geometry", "maxspeed"],
        ),
        (
            {"*speed": ["*0", "90"]},
            293,
            [
                "node/1079045434",
                "way/1001561436",
                "way/1001561437",
                "way/1001561438",
                "way/1019174508",
                "way/1019174509",
                "way/1019646822",
                "way/1024523794",
                "way/1024523795",
                "way/1024523796",
            ],
            2,
            ["geometry", "maxspeed"],
        ),
        (
            {"*speed": "*0", "highway": "primary"},
            356,
            [
                "node/1079045434",
                "way/1001561436",
                "way/1001561437",
                "way/1001561438",
                "way/1019174499",
                "way/1019174508",
                "way/1019174509",
                "way/1019360049",
                "way/1019360050",
                "way/1019646822",
            ],
            3,
            ["geometry", "highway", "maxspeed"],
        ),
        (
            {"*speed": "*0", "highspeed": True},
            293,
            [
                "node/1079045434",
                "way/1001561436",
                "way/1001561437",
                "way/1001561438",
                "way/1019174508",
                "way/1019174509",
                "way/1019646822",
                "way/1024523794",
                "way/1024523795",
                "way/1024523796",
            ],
            2,
            ["geometry", "maxspeed"],
        ),
        (
            {"buildings": {"building": True}},
            1283,
            [
                "relation/11384697",
                "relation/11484092",
                "relation/11484093",
                "relation/11484094",
                "relation/11485520",
                "relation/11538023",
                "relation/11546879",
                "relation/1369192",
                "relation/1369193",
                "relation/1369195",
            ],
            2,
            ["buildings", "geometry"],
        ),
        (
            {"buildings_all": {"building": True}, "buildings_star_all": {"building": "*"}},
            1283,
            [
                "relation/11384697",
                "relation/11484092",
                "relation/11484093",
                "relation/11484094",
                "relation/11485520",
                "relation/11538023",
                "relation/11546879",
                "relation/1369192",
                "relation/1369193",
                "relation/1369195",
            ],
            3,
            ["buildings_all", "buildings_star_all", "geometry"],
        ),
        (
            {"buildings_all": {"building": True}, "buildings_office": {"building": "office"}},
            1283,
            [
                "relation/11384697",
                "relation/11484092",
                "relation/11484093",
                "relation/11484094",
                "relation/11485520",
                "relation/11538023",
                "relation/11546879",
                "relation/1369192",
                "relation/1369193",
                "relation/1369195",
            ],
            3,
            ["buildings_all", "buildings_office", "geometry"],
        ),
        (
            {"english_name": {"name:en": True}, "all_names": {"name:*": True}},
            131,
            [
                "node/10039418191",
                "node/10671507005",
                "node/10674256605",
                "node/11339077004",
                "node/1661205490",
                "node/1704462398",
                "node/1780610146",
                "node/1784106772",
                "node/1790048269",
                "node/2088741979",
            ],
            3,
            ["all_names", "english_name", "geometry"],
        ),
        (
            {"english_name": {"name:en": True}, "e_starting_names": {"name:e*": True}},
            57,
            [
                "node/10039418191",
                "node/10671507005",
                "node/10674256605",
                "node/1661205490",
                "node/1790048269",
                "node/25258130",
                "node/2838235236",
                "node/2838250945",
                "node/4187925032",
                "node/4416171690",
            ],
            3,
            ["e_starting_names", "english_name", "geometry"],
        ),
        (
            {
                "buildings_and_features_with_addr": {"building": True, "addr:*": True},
                "buildings_and_features_with_names": {"building": True, "name:*": True},
            },
            2020,
            [
                "node/10025843517",
                "node/10025852088",
                "node/10025852090",
                "node/10028051243",
                "node/10039418191",
                "node/10060634234",
                "node/10060634235",
                "node/10060634237",
                "node/10060634238",
                "node/10060634239",
            ],
            3,
            ["buildings_and_features_with_addr", "buildings_and_features_with_names", "geometry"],
        ),
        (
            {
                "buildings_without_addr": {"building": True, "addr:*": False},
                "buildings_without_names": {"building": True, "name:*": False},
            },
            282,
            [
                "relation/11484092",
                "relation/11485520",
                "relation/11538023",
                "relation/1369192",
                "relation/1369193",
                "relation/1369195",
                "relation/1484217",
                "relation/16248281",
                "relation/16248282",
                "relation/16248283",
            ],
            3,
            ["buildings_without_addr", "buildings_without_names", "geometry"],
        ),
        (
            {
                "highways_and_ways_with_decimal_speeds": {"highway": True, "*speed": "*0"},
                "railways": {"railway": True},
            },
            3657,
            [
                "node/10068880332",
                "node/10688011472",
                "node/1074584632",
                "node/1074584641",
                "node/1074584644",
                "node/1074584650",
                "node/1074584658",
                "node/1074584679",
                "node/1074584680",
                "node/1074584737",
            ],
            3,
            ["geometry", "highways_and_ways_with_decimal_speeds", "railways"],
        ),
    ],
)
def test_correct_osm_tags_filters(
    osm_tags_filter: Union[OsmTagsFilter, GroupedOsmTagsFilter],
    expected_result_length: int,
    expected_top_10_ids: list[str],
    expected_no_columns: int,
    expected_top_10_columns: list[str],
) -> None:
    """Test proper tags reading with filtering in `PbfFileReader`."""
    file_name = "monaco.osm.pbf"
    features_gdf = PbfFileReader(tags_filter=osm_tags_filter).convert_pbf_to_geodataframe(
        pbf_path=[Path(__file__).parent.parent / "test_files" / file_name],
        ignore_cache=True,
        explode_tags=True,
    )
    assert len(features_gdf) == expected_result_length, (
        f"Mismatched result length ({len(features_gdf)}, {expected_result_length})"
    )
    ut.assertListEqual(list(features_gdf.sort_index().head(10).index), expected_top_10_ids)

    assert len(features_gdf.columns) == expected_no_columns, (
        f"Mismatched columns length ({len(features_gdf.columns)}, {expected_no_columns})"
    )
    ut.assertListEqual(list(features_gdf.columns.sort_values()[:10]), expected_top_10_columns)


@pytest.mark.parametrize(  # type: ignore
    "osm_tags_filter",
    [
        {"name:*": True, "name:en": False},
        {"name:*": ["Monaco", "France"], "name:en": False},
        {"*speed": "*0", "maxspeed": False, "highway": "primary"},
        ({"buildings": {"building": True}, "non-buildings": {"building": False}}),
        ({"buildings_office": {"building": ["office"]}, "non-buildings": {"building": False}}),
        ({"buildings_yes": {"building": "yes"}, "non-buildings": {"building": False}}),
        ({"buildings_all": {"building": "*"}, "non-buildings": {"building": False}}),
    ],
)
def test_incorrect_osm_tags_filters(
    osm_tags_filter: Union[OsmTagsFilter, GroupedOsmTagsFilter],
) -> None:
    """Test wrong tags reading with filtering in `PbfFileReader`."""
    with pytest.raises(ValueError):
        file_name = "monaco.osm.pbf"
        PbfFileReader(tags_filter=osm_tags_filter).convert_pbf_to_geodataframe(
            pbf_path=[Path(__file__).parent.parent / "test_files" / file_name],
            ignore_cache=True,
            explode_tags=False,
        )
